require 'mini_magick'
require 'parallel'

# Resizes an image to provide different versions for multiple
# display resolutions
#
# @param url [String] the image's url
# @param widths [Array<Numeric>] 
#       target sizes to resize to; 
#       only widths smaller than the original width are generated
# @param base_image_path [String] the base image path. Usually assets/images
# @param cache [Jekyll::Cache] 
#       cache object to not resize if the computation was already done
# @return [Array<Numeric>, Numeric] 
#       the generated widths and the original iamge width
def resize_image(url, widths, dest, base_image_path, cache)
    src_path = File.join(base_image_path, url)
    src_mtime = File.new(src_path).mtime
    if cache.key? src_path
        _, _, modified_time = cache[src_path]
        if src_mtime > modified_time
            cache.delete src_path
        end
    end
    # resize images if they aren't already in the cache
    new_widths, w, modified_time = cache.getset(src_path) do
        modified_time = File.new(src_path).mtime
        image = MiniMagick::Image.open(src_path)
        w = image.dimensions[0]
        aspect = image.dimensions[1].to_f / w
        [w, aspect] 
        new_widths = widths.select {|x| x <= w}
        width_to_create = new_widths.map do |width| 
            [width, File.join(dest, base_image_path, "#{width}", url)]
        end
        .select do |width, target| 
            (not File.exists? target) or (File.new(target).mtime < src_mtime)
        end
        Parallel.each(width_to_create) do |width, target|
            image = MiniMagick::Image.open(src_path)
            image.resize "#{width}x#{width*aspect}"
            if not Dir.exists? File.dirname(target)
                FileUtils.mkdir_p File.dirname(target)
            end
            image.write target
        end
        [new_widths, w, modified_time]
    end

    return new_widths, w
end

# Parses the argument list
#
# @param ident [String]
#   the list to parse, e.g., '.someclass key="value"'
# @return [Hash]
#   a hash containing the sorted values
def parse_identifier(ident)
    i = {
        classes: [],
        sizes: "(min-width: 1180px) 1024px, (min-width: 980px) 844px, (min-width: 740px) 640, 100vw"
    }
    if ident.nil?
        return i
    end

    ident.scan(/#[^\s\}]+|\.[^\s\}]+|[^\s\}]+="[^"]+"/) do | match |
        if match.start_with? "#"
            match[0] = ''
            i[:id] = match
        elsif match.start_with? "."
            match[0] = ''
            i[:classes] << match
        else
            parts = match.split '='
            i[parts[0].to_sym] = parts[1].gsub(/^"/, '').gsub(/"$/, '')
        end
    end
    return i
end

def format_image(ident, sizes, original_width, url, caption, title, baseurl, base_image_path)
    srcset = sizes.map {|s| "#{baseurl}/#{base_image_path}/#{s}/#{url} #{s}w"}
    srcset << "#{baseurl}/#{base_image_path}/#{url} #{original_width}w"
    srcset = srcset.join ", "
    classes = "image #{ident[:dimension]} #{ident[:classes].join " "}"
    content = "<div class=\"#{classes}\""
    if not ident[:style].nil?
        content += " style=\"#{ident[:style]}\""
    end
    content += ">"
    content += "<div>"
    srcset_sizes = if not ident[:sizes].nil? and ident[:sizes].length > 0
        ident[:sizes]
    else
        "(min-width: 1180px) 1024px, (min-width: 980px) 844px, (min-width: 740px) 640, 100vw"
    end
    if not caption.nil? and caption != ""
        content += "<figure>
            <img
                    src   = \"#{baseurl}/#{base_image_path}/#{sizes[0]}/#{url}\"
                    alt   = \"#{title}\"
                    sizes = \"#{srcset_sizes}\"
                    srcset= \"#{srcset}\"
                    width = \"100%\"
            />
            <figcaption>#{caption}</figcaption>
        </figure>"
    else
        content += "<img
            src    = \"#{baseurl}/#{base_image_path}/#{sizes[0]}/#{url}\"
            alt    = \"#{title}\"
            sizes  = \"#{srcset_sizes}\"
            srcset = \"#{srcset}\"
        />"
    end
    content += "</div>"
    content += "</div>"
    return content
end

# Formats svg image data by providing the associated image tag
def format_svg_image(url, ident, title, baseurl, base_image_path)
    content = "<img "
    if not ident[:style].nil?
        content += "style=\"#{ident[:style]}\" "
    end
    if not ident[:classes].empty?
        content += "class=\"#{ident[:classes].join " "}\" "
    end
    content += "src    = \"#{baseurl}/#{base_image_path}/#{url}\" "
    content += "alt    = \"#{title}\""
    content += "/>"
end

# def process_page(document, payload)
#     # see https://gist.github.com/mmistakes/77c68fbb07731a456805a7b473f47841
#     doc_ext = document.extname.tr('.', '')
#     cache = Jekyll::Cache.new("jekyll-img-srcset")
#     if payload['site']['markdown_ext'].include? doc_ext
#         content = document.content
#         base_image_path = document.site.config["base_image_path"] || "assets/images"
#         image_regex = /!\[(?<caption>.*?)?\]\((?<url>\S+?)(?<title> .+?)?\)(?<identifier>\{.+?\})?/ 
#         content = content.gsub(image_regex) do | img |
#             img = image_regex.match img
#             if (/\.(jpe?g|png)$/i =~ img["url"]).nil?
#                 "#{base_image_path}/#{img[0]}"
#             elsif img["url"].start_with? "http"
#                 img[0]
#             elsif img["url"].end_with? ".svg"
#                 ident = parse_identifier img[:identifier]
#                 format_svg_image(img["url"], ident, img["title"],
#                     document.site.baseurl)
#             else
#                 if File.exists?(File.join base_image_path, img['url'])
#                     widths, original_width = resize_image(img['url'], 
#                         document.site.config["widths"],
#                         document.site.config["destination"], 
#                         base_image_path, cache)
#                     document.site.keep_files.concat(widths.map do |w|
#                         File.join(base_image_path, "#{w}", img['url'])
#                     end)
#                     ident = parse_identifier img[:identifier]
#                     format_image(ident, widths, original_width,
#                         img["url"], img["caption"], img["title"],
#                         document.site.baseurl, base_image_path)
#                 else
#                     img[0]
#                 end
#             end
#         end
#         # puts content
#         document.content = content
#     end
# end

# Jekyll::Hooks.register [:pages, :posts, :documents], :pre_render do |document, payload|
#     process_page(document, payload)
# end

def check_in_context(context, path)
    parts = path.split(".")
    c = context.dup
    parts.each do | part |
        if not c.is_a?(Array) and c.key? part
            c = c[part]
        else
            return nil
        end
    end
    return c
end

module Jekyll
    class ImgSrcsetTag < Liquid::Tag

      def cache
        @@cache ||= Jekyll::Cache.new("Jekyll::ImgSrcset")
      end
  
      def initialize(tag_name, text, tokens)
        super
        @text = text
      end
  
      def render(context)
        config = context.registers[:site].config
        base_image_path = config["base_image_path"] || "assets/images"
        attrs = {
            classes: [],
        }
        [/(\S+)="([^"]+?)"/, /(\S+)=([^\s\}"][^\s\}]+)/].each do | attr_regex |
            @text.scan(attr_regex) do | key, value |
                if key == "class"
                    context_content = check_in_context context, value
                    if context_content.nil?
                        attrs[:classes] << value
                    else
                        attrs[:classes] << context_content
                    end
                else
                    context_content = check_in_context context, value
                    if context_content.nil?
                        attrs[key.to_sym] = value
                    else
                        attrs[key.to_sym] = context_content
                    end
                end
            end
        end

        if File.exists?(File.join base_image_path, attrs[:src])
            widths, original_width = resize_image(
                attrs[:src], 
                config["widths"],
                config["destination"], base_image_path, cache)
            context.registers[:site].keep_files.concat(widths.map do |w|
                File.join(base_image_path, "#{w}", attrs[:src])
            end)
            format_image(attrs, widths, original_width,
                attrs[:src], attrs[:caption], attrs[:title], 
                context.registers[:site].baseurl, base_image_path)
        else
            puts "image not found: #{File.join base_image_path, attrs[:src]} in page at url #{context["page"]["url"]}"
            ""
        end
      end
    end
end

Liquid::Template.register_tag('imgsrcset', Jekyll::ImgSrcsetTag)